from __future__ import annotations

import os
import re
import sys
from typing import List, Set, Optional

from .constants import PIPE, SPACE_PREFIX, ELBOW, TEE
from .data import Data
from .file import File
from .folder import Folder
from .renderer import Renderer

pattern = re.compile(r"[^\\]")


class DisplayModel:
    def __init__(self, renderer: Renderer):
        self._renderer = renderer
        self._ignore: Set[int] = set()

    def _display_node(self, node: Data) -> str:
        """
        Bookkeeping of nested node depth, node siblings, and order in the queue are
        used to present the FicusSystem in an aesthetic way.
        """

        parts = [PIPE + SPACE_PREFIX for _ in range(node.level)]
        for index in self._ignore:
            if len(parts) > index - 1:
                parts[index - 1] = " " + SPACE_PREFIX

        if parts:
            parts[-1] = ELBOW if node.last is True else TEE

        is_file = isinstance(node, File)
        file_open = self._renderer.file_open if is_file else self._renderer.folder_open
        file_close = (
            self._renderer.file_close if is_file else self._renderer.folder_close
        )

        # checking for Folder type
        end = "\\" if not is_file else ""

        return f'{"".join(parts)}{file_open}{node.name}{file_close}{end}'

    def _display_header(self, header: str) -> int:
        """Writes the CWD to stdout with forward slashes and its length."""
        ONE_PAST_FOLDER = 1
        parts = header.split(os.sep)
        if len(parts) > 1:
            header = f"\\".join(parts[:-1])
            sys.stdout.write(f"{header}\\\n")
            return header[:-1].rfind("\\") + ONE_PAST_FOLDER

        # when root is passed in
        return 0

    def display(self, node: Data, root: str) -> None:
        """Prints the directory structure to stdout."""
        sys.stdout.write(self._renderer.doc_open + "\n")
        prefix: Optional[int] = None
        header_length = self._display_header(root)

        node.last = True
        q: List[Data] = [node]

        self._ignore = {i for i in range(node.level)}
        while q:
            node = q.pop()
            if node.last is False:
                if node.level in self._ignore:
                    self._ignore.remove(node.level)

            line = self._display_node(node)
            if prefix is None:
                # This needs to happen only once and applied
                # thereafter to each subsequent line.
                prefix = len(line) - len(line.lstrip())

            sys.stdout.write(header_length * " " + line[prefix:] + "\n")
            if node.last is True:
                # track nodes without children.
                self._ignore.add(node.level)

            if isinstance(node, Folder):
                q += node.contents()

        sys.stdout.write(self._renderer.doc_close + "\n")



import os.path
from typing import Set, Optional

from .displaymodel import DisplayModel
from .fictusexception import FictusException
from .file import File
from .folder import Folder
from .renderer import defaultRenderer


class FictusFileSystem:
    """A FictusFileSystem represent a fake file system (FFS)."""

    DEFAULT_ROOT_NAME = "\\"

    def __init__(self, name=DEFAULT_ROOT_NAME) -> None:
        self.level: int = 0
        self.root: Folder = Folder(name)
        self.current: Folder = self.root
        self._display_model: Optional[DisplayModel] = None

    def set_display_model(self, display) -> None:
        self._display_model = display

    def display(self) -> None:
        if self._display_model is None:
            self._display_model = DisplayModel(defaultRenderer)

        self._display_model.display(self.current, self.cwd())

    @staticmethod
    def _normalize(path: str) -> str:
        return os.path.normpath(path.replace("\\", "/"))

    def mkdir(self, path: str) -> None:
        """Takes a string of a normalized relative to cwd and adds the directories
        one at a time."""
        if not path:
            raise FictusException("A Folder must contain a non-empty string.")

        # hold onto the current directory
        current = self.current
        current_level = self.level

        normalized_path = self._normalize(path)
        if normalized_path.startswith(os.sep):
            self._to_root()

        visited = {d.name: d for d in self.current.folders()}

        for part in normalized_path.split(os.sep):
            if not part:
                continue

            if part not in visited:
                visited[part] = Folder(part)
                self.current.folder(visited[part])

            self.cd(visited[part].name)
            visited = {d.name: d for d in self.current.folders()}

        # return to starting directory
        self.current = current
        self.level = current_level

    def mkfile(self, *files: str) -> None:
        """Takes one or more filenames and adds them to the cwd."""
        visited: Set[str] = {f.name for f in self.current.files()}
        for file in files:
            if not file:
                raise FictusException("A File must contain a non-empty string.")

            if file not in visited:
                visited.add(file)
                self.current.file(File(file))

    def rename(self, old: str, new: str) -> None:
        """Renames a File or Folder based on its name."""
        for content in self.current.contents():
            if content.name == old:
                content.name = new
                break

    def cwd(self):
        """Prints the current working directory."""
        r = []
        visited = set()
        q = [self.current]
        while q:
            n = q.pop()
            if n.name is not None:
                r.append(n.name)
            visited.add(n)
            if n.parent and n.parent not in visited:
                q.append(n.parent)

        return f"{os.sep}".join(r[::-1])

    def _to_root(self) -> None:
        self.current = self.root

    def cd(self, path: str) -> None:
        """Takes a string of a normalized relative to cwd and changes the current"""

        normalized_path = self._normalize(path)
        if normalized_path.startswith(os.sep):
            self._to_root()

        for index, part in enumerate(normalized_path.split(os.sep)):
            if not part:
                continue

            if index == 0 and part == self.root.name:
                self._to_root()
                continue

            if part == "..":
                if self.current.parent is None:
                    raise FictusException(
                        f"Folder not accessible; Attempted to move {part} from {self.cwd()}."
                    )
                self.current = self.current.parent
                self.level = self.current.level
            else:
                hm = {f.name: f for f in self.current.folders()}
                if part not in hm:
                    raise FictusException(
                        f"Folder not accessible; Attempted to move {part} from {self.cwd()}."
                    )
                self.current = hm[part]
                self.level = self.current.level

        return None
